// spimem - SPI memory interface
//
// Protocol is:
//   - as many bytes as required to send the address, big-endian MSB-first
//   - optional stride word, see below
//   - mode byte (MSB 0 = read, MSB 1 = write)
//   - data bytes, big-endian if DATAW > 1
//
// Spimem can be configured to write with a stride, to allow writing e.g.
// rectangular regions in a video RAM. To do this, set parameter LINEW to
// something other than 0. Then, the optional stride word must be included in
// the protocol, and is equal in width to an address word. After every
// `stride` words, the target address will skip forward to the next line.
// `stride` may be zero to never skip.

module spimem(
	// SPI port. Currently always mode 0 MSB-first
	input      sck,
	input      copi,
	output     cipo,
	input      cs,   // active high

	// Address for writes and reads
	output reg [ADDR_TOP:0] addr,

	// Write port. Once we asserts, it will remain for half the data word.
	output reg [DATA_TOP:0] wdata,
	output reg              we,

	// Read port. Connect rdata to 0 if not using. Once re asserts or addr
	// changes while re is asserted, you have 6 scks to get the new data
	// out.
	input      [DATA_TOP:0] rdata,
	output reg              re,

	// Asynchronous reset
	input      rst
);

parameter  ADDRW  = 1; // Configurable: width of the address, in bytes
parameter  DATAW  = 1; // Configurable: width of the data words, in bytes
parameter  LINEW  = 0; // Configurable: a line is LINEW words long

localparam ADDR_TOP = ADDRW * 8 - 1;
localparam DATA_TOP = DATAW * 8 - 1;

localparam SHIFTW    = (ADDRW > DATAW) ? ADDRW : DATAW;
localparam SHIFT_TOP = SHIFTW * 8 - 1;

localparam COUNTER_TOP = $clog2(SHIFTW * 8) - 1;

localparam S_ADDR = 0, S_STRIDE = 1, S_MODE = 2, S_READ = 3, S_WRITE = 4,
	S_MAX = 4;

reg  [S_MAX:0]       r_state; // state machine, coded one-hot
reg  [SHIFT_TOP:0]   r_shift; // SPI shift register, for both read and write
reg  [COUNTER_TOP:0] r_count; // bit counter to track position in word

reg  [ADDR_TOP:0]    r_line_left; // Lefthand address of the next line
reg  [ADDR_TOP:0]    r_stride;    // Stride, for rectangular writes
reg  [ADDR_TOP:0]    r_strcnt;    // Stride counter

wire [SHIFT_TOP:0]   next_shift;  // next value for r_shift

// cipo and copi always connect to r_shift for simplicity
assign next_shift = { r_shift[SHIFT_TOP - 1 : 0], copi };
assign cipo = r_shift[DATA_TOP];

task advance_addr;
	if (LINEW && r_strcnt == r_stride - 1) begin
		addr <= r_line_left;
		r_line_left <= r_line_left + LINEW;
		r_strcnt <= 0;
	end else begin
		addr <= addr + 1;
		r_strcnt <= r_strcnt + 1;
	end
endtask

always @(posedge(cs), posedge(rst)) begin
	r_shift <= 0;
	r_state <= (4'b0001 << S_ADDR);
	r_count <= 0;
	r_stride <= 0;
	r_strcnt <= 0;
	we <= 0;
	re <= 0;
	addr <= 0;
	wdata <= 0;
end

always @(posedge(sck)) begin: sm_blk
	integer count_top;

	if (cs) begin
		r_state <= 0;
		r_shift <= next_shift;

		case (1'b1)
		r_state[S_ADDR]:
		begin
			count_top = ADDR_TOP;
			if (r_count == ADDR_TOP) begin
				r_state[LINEW ? S_STRIDE : S_MODE] <= 1'b1;
				addr <= next_shift[ADDR_TOP:0];
			end else begin
				r_state[S_ADDR] <= 1'b1;
			end
		end
		r_state[S_STRIDE]:
		begin
			count_top = ADDR_TOP;
			if (r_count == ADDR_TOP) begin
				r_state[S_MODE] <= 1'b1;
				r_stride <= next_shift[ADDR_TOP:0];
				r_line_left <= addr + LINEW;
			end else begin
				r_state[S_STRIDE] <= 1'b1;
			end
		end
		r_state[S_MODE]:
		begin
			count_top = 7;
			if (r_count == 0) begin
				re <= ~copi;
				r_state[S_MODE] <= 1'b1;
			end else if (r_count == 7) begin
				r_state[re ? S_READ : S_WRITE] <= 1'b1;
			end else begin
				r_state[S_MODE] <= 1'b1;
			end
		end
		r_state[S_READ]:
		begin
			count_top = DATA_TOP;
			r_state[S_READ] <= 1'b1;
			if (r_count == 0) begin
				r_shift <= rdata;
				advance_addr();
			end else begin
			end
		end
		r_state[S_WRITE]:
		begin
			count_top = DATA_TOP;
			r_state[S_WRITE] <= 1'b1;
			if (r_count == DATA_TOP/2) begin
				// Terminate the write halfway to the next
				we <= 0;
				if (we) begin
					advance_addr();
				end
			end else if (r_count == DATA_TOP) begin
				// last write bit
				wdata <= next_shift[DATA_TOP:0];
				we <= 1;
			end else begin
			end
		end
		endcase

		r_count <= (r_count == count_top) ? 0 : r_count + 1;
	end
end

endmodule
