// spimux - SPI multiplexer
//
// Eats the first byte of an SPI transaction and uses it to select
// a downstream mailbox.
//
// The most significant bits encode the mailbox, and this command word is also
// emitted, so the remaining bits can be used elsewhere.
//
// Assumes mode 0, MSB-first.
//
// sck is the clock domain for this part and should be routed to all
// downstream mailboxes as well. All downstream mailbox IOs are valid for port
// n when mcs[n] asserts.
//
// Uses active HIGH chip select per general convention of all internal logic
// being active high.

module spimux(
	input      sck,
	input      copi,
	output reg cipo,
	input      cs,

	output reg [n_mailboxes - 1:0] mcopi,
	output reg [n_mailboxes - 1:0] mcs,
	input      [n_mailboxes - 1:0] mcipo,

	output reg [7:0] cmd
);

integer i;

parameter n_mailboxes = 2;
// Top bit of the mailbox ID
localparam mb_top_bit = $clog2(n_mailboxes) - 1;
// Bottom bit of the mailbox ID within the command byte
localparam mb_cmd_bot_bit = 7 - $clog2(n_mailboxes) + 1;

reg [7:0] r_shift;
reg [3:0] r_state; // 0-7 are command bits, then stick at 8 until next cs

wire [7:0] next_shift;
wire [mb_top_bit:0] mailbox_sel;

assign mailbox_sel = cmd[7 : mb_cmd_bot_bit];
// assign next_shift  = { copi, r_shift[7:1] }; // LSB-first
assign next_shift  = { r_shift[6:0], copi };

initial begin
	cmd <= 0;
	r_shift <= 0;
end

always @(posedge(cs)) begin
	r_state <= 4'b0;
end

always @(negedge(cs)) begin
	mcs <= 0;
end

always @(posedge(sck)) begin
	if (r_state == 4'b111) begin
		cmd <= next_shift;
		mcs[ next_shift[7 : mb_cmd_bot_bit] ] <= 1;
	end

	if (r_state[3] == 0) begin
		r_state <= r_state + 1;
	end

	r_shift <= next_shift;
end

// Just route copi to all the mailboxes
always @(copi) begin
	for (i = 0; i < n_mailboxes; i = i + 1) begin
		mcopi[i] = copi;
	end
end

// cipo mux
always @(mailbox_sel or mcipo) begin
	cipo = mcipo[mailbox_sel];
end

endmodule
