// /opt/microchip/xc8/v2.32/bin/xc8-cc -mcpu=attiny806 -O2 supervisor.c
#define F_CPU 1000000

#include <avr/io.h>
#include <util/delay.h>
#include <stdbool.h>

FUSES = {
	.WDTCFG = 0,
	.BODCFG = BOD_LVL_BODLEVEL2_gc, // 2.6V
	.OSCCFG = FREQSEL_16MHZ_gc,
	.SYSCFG0 = CRCSRC_NOCRC_gc | RSTPINCFG_UPDI_gc,
	.SYSCFG1 = SUT_4MS_gc,
	.APPEND = 0,
	.BOOTEND = 0,
};

#define PORT_PWREN PORTA
#define PIN_PWREN_bp 7u

#define PORT_EN_MOST PORTB
#define PIN_EN_MOST_bp 0u

#define PORT_EN_VCORE PORTB
#define PIN_EN_VCORE_bp 1u

#define PORT_CPU_nRST PORTB
#define PIN_CPU_nRST_bp 3u

#define PORT_CPU_nSHDN PORTB
#define PIN_CPU_nSHDN_bp 4u

#define PORT_nEN_FPGA_VCCAUX PORTB
#define PIN_nEN_FPGA_VCCAUX_bp 5u

#define PORT_nSEL_5V PORTC
#define PIN_nSEL_5V_bp 0u

#define PIN_VIN_AIN 1
#define PIN_5V0_AIN 2
#define PIN_3V3_AIN 3
#define PIN_2V5_AIN 4
#define PIN_1V8_AIN 5
#define PIN_1V2_AIN 6
#define PORT_ANALOG PORTA
#define PINS_ANALOG_gm 0x7Eu

static void _check_and_drive_sel5v(void);
static void _wait_pwrgd_1v2(void);
static void _wait_pwrgd(void);
static void _do_startup(void);
static void _do_shutdown(void);

int main(void)
{
	_PROTECTED_WRITE(CLKCTRL.MCLKCTRLA, CLKCTRL_CLKSEL_OSC20M_gc);
	_PROTECTED_WRITE(CLKCTRL.MCLKCTRLB, CLKCTRL_PDIV_16X_gc | CLKCTRL_PEN_bm);

	PORT_PWREN.DIR &= ~(1u << PIN_PWREN_bp);

	PORT_EN_MOST.OUT &= ~(1u << PIN_EN_MOST_bp); // init 0
	PORT_EN_MOST.DIR |= (1u << PIN_EN_MOST_bp);

	PORT_EN_VCORE.OUT &= ~(1u << PIN_EN_VCORE_bp); // init 0
	PORT_EN_VCORE.DIR |= (1u << PIN_EN_VCORE_bp);

	PORT_CPU_nRST.OUT &= ~(1u << PIN_CPU_nRST_bp); // init 0
	PORT_CPU_nRST.DIR |= (1u << PIN_CPU_nRST_bp);

	PORT_CPU_nSHDN.OUT &= ~(1u << PIN_CPU_nSHDN_bp); // init 0
	PORT_CPU_nSHDN.DIR |= (1u << PIN_CPU_nSHDN_bp);

	PORT_nEN_FPGA_VCCAUX.OUT |= (1u << PIN_nEN_FPGA_VCCAUX_bp); // init 1
	PORT_nEN_FPGA_VCCAUX.DIR |= (1u << PIN_nEN_FPGA_VCCAUX_bp); // init 1

	// Shut off the input buffers on analog inputs
	uint8_t volatile * pinctrl = &PORT_ANALOG.PIN0CTRL;
	uint8_t pin_bit = 1;
	for (uint8_t i = 0; i < 8; i++, pin_bit <<= 1, pinctrl++)
	{
		if (PINS_ANALOG_gm & pin_bit)
		{
			*pinctrl = PORT_ISC_INPUT_DISABLE_gc;
		}
	}

	bool state = false;

	for (;;)
	{
		bool const pwren = (PORT_PWREN.IN & (1u << PIN_PWREN_bp));

		if (pwren && !state)
		{
			state = true;
			_do_startup();
		}
		else if (!pwren && state)
		{
			state = false;
			_do_shutdown();
		}
	}
}

static void _do_startup(void)
{
	_delay_ms(1);

	_check_and_drive_sel5v();

	PORT_nEN_FPGA_VCCAUX.OUT |= 1u << PIN_nEN_FPGA_VCCAUX_bp;
	PORT_EN_VCORE.OUT |= (1u << PIN_EN_VCORE_bp);

	_wait_pwrgd_1v2();
	_delay_ms(1);

	PORT_EN_MOST.OUT |= (1u << PIN_EN_MOST_bp);
	_wait_pwrgd();
	_delay_ms(1);

	PORT_nEN_FPGA_VCCAUX.OUT &= ~(1u << PIN_nEN_FPGA_VCCAUX_bp);

	PORT_CPU_nSHDN.DIR &= ~(1u << PIN_CPU_nSHDN_bp);
	PORT_CPU_nRST.DIR &= ~(1u << PIN_CPU_nRST_bp);
}

static void _do_shutdown(void)
{
	PORT_CPU_nSHDN.DIR |= (1u << PIN_CPU_nSHDN_bp);
	_delay_ms(250);
	PORT_CPU_nRST.DIR |= (1u << PIN_CPU_nRST_bp);
	_delay_ms(1);
	PORT_EN_MOST.OUT &= ~(1u << PIN_EN_MOST_bp);
	PORT_EN_VCORE.OUT &= ~(1u << PIN_EN_VCORE_bp);
}

static void _check_and_drive_sel5v(void)
{
}

static void _wait_pwrgd_1v2(void)
{
}

static void _wait_pwrgd(void)
{
}
