#!/usr/bin/python3

import argparse
import collections
import csv
import itertools
import math
import os
import re
import shutil
import sys
import time

import colorama
from recordclass import recordclass

VERBOSE = 2
ARGS = None
OUTDIR = None
NAMES = {}
EXPORTS = {}

Bomline = recordclass("Bomline", ("refs", "dist", "item", "qty",
    "custref"))
PartDbEntry = recordclass("PartDbEntry", ("name", "stock", "dist",
    "dpn", "restock", "pkg"))
StockUseEntry = recordclass("StockUseEntry", ("name", "qty", "refs"))

GERB_MATCHERS = [
    ("F_Fab",   re.compile(r'.*-F[_.]Fab\.gbr$', re.I)),
    ("F_CrtYd", re.compile(r'.*-F[_.]CrtYd\.gbr$', re.I)),
    ("F_Paste", re.compile(r'.*-F[_.]Paste\.(gtp|gbr)$|.*\.gtp$', re.I)),
    ("F_SilkS", re.compile(r'.*-F[_.](SilkS|Silkscreen)\.(gto|gbr)$|.*\.gto$|to$', re.I)),
    ("F_Mask",  re.compile(r'.*-F[_.]Mask\.(gts|gbr)$|.*\.gts$|ts$', re.I)),
    ("F_Cu",    re.compile(r'.*-F[_.]Cu\.(gtl|gbr)$|.*\.gtl$|.*\.gl1$|.*\.g1l$|tl$', re.I)),
    ("In1_Cu",  re.compile(r'.*-In1[_.]Cu\.gbr$|.*\.gl2$|.*\.g2l|.*\.g2$|.*\.gp1|l2$', re.I)),
    ("In2_Cu",  re.compile(r'.*-In2[_.]Cu\.gbr$|.*\.gl3$|.*\.g3l|.*\.g3$|.*\.gp2|l3$', re.I)),
    ("In3_Cu",  re.compile(r'.*-In3[_.]Cu\.gbr$|.*\.gl4$|.*\.g4l|.*\.g4$|l4$', re.I)),
    ("In4_Cu",  re.compile(r'.*-In4[_.]Cu\.gbr$|.*\.gl5$|.*\.g5l|.*\.g5$|l5$', re.I)),
    ("B_Cu",    re.compile(r'.*-B[_.]Cu\.(gbl|gbr)$|.*\.gbl$|bl$', re.I)),
    ("B_Mask",  re.compile(r'.*-B[_.]Mask\.(gbs|gbr)$|.*\.gbs$|bs$', re.I)),
    ("B_SilkS", re.compile(r'.*-B[_.](SilkS|Silkscreen)\.(gbo|gbr)$|.*\.gbo$|bo$', re.I)),
    ("B_Paste", re.compile(r'.*-B[_.]Paste\.(gbp|gbr)$|.*\.gbp$', re.I)),
    ("B_CrtYd", re.compile(r'.*-B[_.]CrtYd\.gbr$', re.I)),
    ("B_Fab",   re.compile(r'.*-B[_.]Fab\.gbr$', re.I)),
    ("Eco1",    re.compile(r'.*-Eco1[_.]User\.gbr$', re.I)),
    ("Eco2",    re.compile(r'.*-Eco2[_.]User\.gbr$', re.I)),
    ("Edge_Cuts", re.compile(r'.*-Edge[_.]Cuts\.gbr$|.*\.gko|.*\.gml|.*\.gm1|ko$', re.I)),
    ("PTH",     re.compile(r'((?!-NPTH).)*\.(drl|xln|txt)$|drl$', re.I)),
    ("NPTH",    re.compile(r'.*-NPTH\.drl$', re.I)),
]

FAB_GERBER_NAMES = {
    "F_Cu":      ".GTL",
    "In1_Cu":    ".GL2",
    "In2_Cu":    ".GL3",
    "In3_Cu":    ".GL4",
    "In4_Cu":    ".GL5",
    "B_Cu":      ".GBL",

    "F_SilkS":   ".GTO",
    "B_SilkS":   ".GBO",

    "F_Mask":    ".GTS",
    "B_Mask":    ".GBS",

    "F_Paste":   ".GTP",
    "B_Paste":   ".GBP",

    "Edge_Cuts": ".GML",
    "PTH":       "-PTH.TXT",
    "NPTH":      "-NPTH.TXT",
}

def verbose1(arg0, *args, **kwargs):
    if VERBOSE >= 1:
        print(str(arg0), *args, **kwargs,
            end=colorama.Style.RESET_ALL + "\n")

def verbose2(arg0, *args, **kwargs):
    if VERBOSE >= 1:
        print(colorama.Fore.CYAN + str(arg0), *args, **kwargs,
            end=colorama.Style.RESET_ALL + "\n")

def die(*args, **kwargs):
    print(colorama.Fore.RED + "error:", *args, **kwargs, file=sys.stderr,
        end=colorama.Style.RESET_ALL + "\n")
    sys.exit(1)

def warn(*args, **kwargs):
    print(colorama.Fore.YELLOW + "warning:", *args, **kwargs, file=sys.stderr,
        end=colorama.Style.RESET_ALL + "\n")

def identify_gerber(fn):
    """Takes a filename and identifies which Gerber layer it is.

    Return values:

    "F_Fab", "F_CrtYd", "F_Paste", "F_SilkS", "F_Mask", "F_Cu", "In1_Cu", "In2_Cu", "B_Cu", "B_Mask",
    "B_SilkS", "B_Paste", "B_CrtYd", "B_Fab", "Eco1", "Eco2", "Edge_Cuts", "PTH", "NPTH", None
    """

    for name, matcher in GERB_MATCHERS:
        if matcher.match(fn) is not None:
            return name
def main():
    global ARGS
    global OUTDIR
    ARGS = get_args()

    OUTDIR = os.path.join("orders", time.strftime("%Y%m%d"))
    verbose1("  Output to:", OUTDIR)

    try:
        os.makedirs(OUTDIR)
    except FileExistsError:
        warn("output path exists")

    partsdb = load_partsdb()
    bom = []

    for i in ARGS.projects:
        check_project(i)
        pack_gerbers(i)

        bom.extend(load_bom(i))

    bom = fold_bom(bom)
    restock, stockuse = resolve_bom(bom, partsdb)
    emit_dist_boms(itertools.chain(bom, restock))

def get_args():
    ap = argparse.ArgumentParser(description="Generate an order package")
    ap.add_argument("projects", help="KiCad projects to read", nargs="+")
    ap.add_argument("--qty", type=int, help="Quantity multiplier", default=1)

    return ap.parse_args()

def check_project(project):
    global EXPORTS

    verbose1("Checking project:", project)

    NAMES[project] = os.path.basename(project)
    verbose2("  Project name is:", NAMES[project])

    EXPORTS[project] = os.path.join(project, "exports")
    if not os.path.isdir(EXPORTS[project]):
        die("exports path missing:", EXPORTS[project])

def pack_gerbers(project):
    gerbdir = os.path.join(OUTDIR, NAMES[project] + "-gerbers")
    verbose1("Packing gerbers into", gerbdir)
    os.makedirs(gerbdir, exist_ok=True)

    for ent in os.scandir(EXPORTS[project]):
        if not ent.is_file():
            continue

        gname = identify_gerber(ent.name)

        if gname is None:
            # This is normal
            verbose2("    gerbers: skipping unknown file:", ent.path)
            continue

        fabname = FAB_GERBER_NAMES.get(gname)
        if fabname is None:
            warn("gerbers: recognized file but have no name for it:", ent.path)
            continue

        newpath = os.path.join(gerbdir, NAMES[project] + fabname)
        verbose2("   ", ent.path, "->", newpath)
        shutil.copyfile(ent.path, newpath)

    verbose1("Creating archive:", gerbdir + ".zip")
    shutil.make_archive(gerbdir, "zip",
        root_dir=OUTDIR,
        base_dir=NAMES[project] + "-gerbers",
    )

def load_bom(project):
    """Load the project BOM and return a list of Bomline"""

    bom_fn = None
    for ent in os.scandir(EXPORTS[project]):
        if not ent.is_file():
            continue
        if ent.path.endswith(".csv"):
            bom_fn = ent.path
            verbose1("Loading BOM from", bom_fn)
            break
    else:
        warn("No BOM found")
        return []

    bom = []
    name = NAMES[project]

    with open(bom_fn) as f:
        reader = csv.reader(f, skipinitialspace=True)
        head = next(reader)

        try:
            col_reference = head.index("Reference")
            col_bom = head.index("BOM")
        except ValueError:
            warn("empty BOM:", project)
            return []

        for row in reader:
            try:
                reference = row[col_reference]
            except IndexError:
                verbose2("skipping empty line")

            try:
                bomline = row[col_bom]
            except IndexError:
                warn("no BOM line for", reference)
                continue
            if not bomline.strip():
                warn("no BOM line for", reference)
                continue

            splits = [i.strip() for i in bomline.split(";")]

            for bomline in splits:
                qty, _, single = bomline.partition("#")
                if _ == "#":
                    bomline = single
                    qty = int(qty)
                else:
                    qty = 1

                distrib, _, item = bomline.partition(":")

                if _ == "":
                    bom.append(Bomline([reference], "stock", bomline,
                        qty=ARGS.qty * qty, custref=name))
                else:
                    bom.append(Bomline([reference], distrib, item,
                        qty=ARGS.qty * qty, custref=name))

    return bom

def load_partsdb():
    """Load the parts database and return a list of PartDbEntry"""

    db = []

    verbose1("Loading part database")
    with open("partsdb.csv") as f:
        reader = csv.reader(f, skipinitialspace=True)
        head = next(reader)

        col_name = head.index("Name")
        col_stock = head.index("Stock")
        col_restock = head.index("Restock")
        col_pkg = head.index("Pkg")
        col_dist = head.index("Dist")
        col_dpn = head.index("DPN")

        if (col_name < 0 or col_stock < 0 or col_restock < 0 or col_pkg < 0 or
                col_dist < 0 or col_dpn < 0):
            die("invalid partsdb.csv: expected Name, Stock, Restock, Pkg, Dist, DPN")

        for row in reader:
            name = row[col_name]
            stock = int(row[col_stock])
            restock = int(row[col_restock])
            pkg = int(row[col_pkg])
            dist = row[col_dist]
            dpn = row[col_dpn]
            db.append(PartDbEntry(name=name, stock=stock, dist=dist, dpn=dpn,
                restock=restock, pkg=pkg))

    return db

def fold_bom(bom):
    """Fold up a BOM so each entry is unique. Each entry.ref will now be a
    list of references. Returns the new, folded BOM"""

    bom.sort(key = lambda e: (e.dist, e.item))

    new = []

    for ent in bom:
        if new and (ent.dist, ent.item) == (new[-1].dist, new[-1].item):
            new[-1].refs.extend(ent.refs)
            new[-1].custref.add(ent.custref)
        else:
            new.append(Bomline(refs=ent.refs[:], dist=ent.dist, item=ent.item,
                qty=ent.qty, custref=set([ent.custref])))

    for ent in new:
        ent.custref = ";".join(ent.custref)
    return new

def resolve_bom(bom, partsdb):
    """Resolve stock entries in the BOM according to the parts database.

    Returns two lists: restock, stockuse

    restock: [Bomline] for parts that need to be restocked
    stockuse: [StockUseEntry] for parts being used
    """

    restock = []
    stockuse = []

    for entry in bom:
        if entry.dist != "stock":
            continue

        qty = len(entry.refs) * entry.qty

        # Find the item in stock (alternate with lowest sufficient qty)
        matches = [i for i in partsdb if i.name == entry.item]
        if not matches:
            die("Missing stock entry:", entry.item)

        matches.sort(key = lambda e: e.stock)

        stock_source = None
        for i in matches:
            if i.stock >= qty:
                stock_source = i
                break
        else:
            stock_source = resolve_alts(entry.item, matches)

        if stock_source.stock <= qty:
            if stock_source.restock <= qty:
                restock_qty = stock_source.restock + qty
            else:
                restock_qty = stock_source.restock

            restock_qty = math.ceil(restock_qty / stock_source.pkg)

            restock.append(Bomline(refs=["restock"], dist=stock_source.dist,
                item=stock_source.dpn, qty=restock_qty,
                custref="Stock:"+ stock_source.name))

        stockuse.append(StockUseEntry(entry.item, qty, entry.refs))

    return restock, stockuse

def resolve_alts(item, matches):
    if len(matches) == 1:
        print("Not enough stock of", item)
        return matches[0]

    print()
    print("Not enough stock of", item)

    for n, i in enumerate(matches):
        print(f"  Alt {n+1}: {i}")

    while True:
        sel = input("Which to buy? ")
        try:
            sel = int(sel)
        except ValueError:
            continue
        if sel < 1 or sel > len(matches):
            continue
        return matches[sel - 1]

def emit_dist_boms(items):
    """Emit BOMs for each distributor. Entries in `bom` will be given a
    customer reference listing designators; entries in `restock` will be given
    a customer reference of RESTOCK."""

    items = list(items)
    dists = set(i.dist for i in items)

    for i in dists:
        emit_dist_bom(items, i)

def emit_dist_bom(items, dist):
    """Emit one BOM for a single distributor (see emit_dist_boms)"""

    items = list(i for i in items if i.dist == dist)

    target = os.path.join(OUTDIR, f"bom-{dist}.csv")
    verbose1("Generating BOM:", target)

    with open(target, "w") as f:
        print("Qty,PN,CustRef", file=f)

        for i in items:
            qty = len(i.refs) * i.qty
            pn = i.item
            custref = i.custref
            print(f"{qty},{pn},{custref}", file=f)

def format_ref_list(refs, collate):
    """Taking a list of refs, format a more readable string."""

    # Split each ref into a tuple of sortables, then sort them. Collate ranges.

    single_number_re = re.compile(
        r'([^\d]+)(\d+)$'
    )

    two_number_re = re.compile(
        r'([^\d]+)(\d+)\.(\d+)$'
    )

    SingleNumber = collections.namedtuple("SingleNumber", ["kind", "num"])
    TwoNumber = collections.namedtuple("TwoNumber", ["kind", "num1", "num"])
    Generic = collections.namedtuple("Generic", ["wtf_even"])

    refs_parsed = []

    for ref in refs:
        single_number_m = single_number_re.match(ref)
        two_number_m = two_number_re.match(ref)

        if single_number_m is not None:
            refs_parsed.append(
                SingleNumber(single_number_m.group(1), [int(single_number_m.group(2))])
            )
        elif two_number_m is not None:
            refs_parsed.append(
                TwoNumber(
                    two_number_m.group(1),
                    int(two_number_m.group(2)),
                    [int(two_number_m.group(3))]
                )
            )
        else:
            refs_parsed.append(Generic(ref))

    refs_parsed.sort()

    if collate:
        collates = refs_parsed[:1]

        for ref in refs_parsed[1:]:
            last = collates[-1]
            if type(ref) is not type(last) or isinstance(ref, Generic):
                collates.append(ref)
                continue
            if ref.num[0] == (last.num[-1] + 1):
                last.num.extend(ref.num)
            else:
                collates.append(ref)
    else:
        collates = refs_parsed

    out = []

    for ref in collates:
        if isinstance(ref, Generic):
            out.append(ref.wtf_even)
        elif isinstance(ref, (TwoNumber, SingleNumber)):
            if len(ref.num) == 1:
                num = ref.num[0]
            else:
                num = f"{ref.num[0]}-{ref.num[-1]}"

            if isinstance(ref, TwoNumber):
                out.append(f"{ref.kind}{ref.num1}.{num}")
            else:
                out.append(f"{ref.kind}{num}")
        else:
            raise Exception(r"¯\_(ツ)_/¯")

    return "+".join(out)

if __name__ == "__main__":
    main()
